function bemt(OpRot, Mod)
    % BEMT Solves BEMT equations using the chosen solver.
    %   This is function is at the core of Rotare. It implements different ways to solve the
    %   Blade Element Momentum equations and returns the Rotor object with its performances.
    % Note:
    %   It solves each rotor separately, one at a time. Before solving the BEMT equations for a
    %   Rotor, this function will determine the true external velocity distribution by taking into
    %   account the output flow of the upstream rotor.
    % -----
    %
    % Syntax:
    %   BEMT(OpRot, Mod) Calculates the performances of each OpRot object with the Blade Element
    %                    Momentum Theory, using the solvers and models defined in Mod
    %
    % Inputs:
    %   OpRot : Operating rotor object(s) fully defined
    %   Mod   : Model and numerical parameters
    %
    % Output:
    %   OperRotor object(s) with calculated performances
    %
    % See also: rotare, leishman, indfact, indvel, stahlhut.
    %
    % <a href="https://gitlab.uliege.be/rotare/documentation">Complete documentation (online)</a>

    % ----------------------------------------------------------------------------------------------
    % TODO: Implement oblique flows
    % ----------------------------------------------------------------------------------------------
    % (c) Copyright 2022-2023 University of Liege
    % Author: Thomas Lambert <t.lambert@uliege.be>
    % ULiege - Aeroelasticity and Experimental Aerodynamics
    % MIT License
    % Repo: https://gitlab.uliege.be/rotare/rotare
    % Docs: https://gitlab.uliege.be/rotare/documentation
    % Issues: https://gitlab.uliege.be/rotare/rotare/-/issues
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    % First rotor always calculated as if isolated
    bemtsinglerot(OpRot(1), Mod);
    
    % Deriving parameters of the second rotor to neutralize the overall torque 
    if length(OpRot) == 2 % The following torque cancellation method is only valid for N_ROTORS = 2
        
        % Solve iteratively
        isConverged = 0;
        loopCount = 0;
        
        % Parameters to vary to make both rotor torques coincide
        strategy = Mod.coax.strategy;
            if ~strcmp(strategy, 'rpm') && ~strcmp(strategy, 'coll')
                error('bemt: the strategy must be''rpm'' or''coll''.');
            end
            
        % Lower and Upper bounds for bissection method
        if strcmp(strategy, 'rpm')
            OpRot_l = 1/10*OpRot(1).Op.(strategy);
            OpRot_u = 10*OpRot(1).Op.(strategy);
        else
            OpRot_l = -10*OpRot(1).Op.(strategy);
            OpRot_u = 10*OpRot(1).Op.(strategy);
        end
        
        while (~isConverged) && loopCount <= Mod.Num.maxIter
            
            OpRot(2).Op.(strategy) = (OpRot_l + OpRot_u)/2; %Midpoint
            
            OpRot(2).ElPerf.updateupstreamvel(OpRot(1), Mod); % Update upstream velocity
            bemtsinglerot(OpRot(2), Mod); % Calculate BEMT as usual, with new upstream vel distribution
            
            % Convergence criterion 
            if abs(OpRot(1).torque - OpRot(2).torque) < Mod.Num.convCrit
                isConverged = 1;
            else
                % Update the bounds based on the torque
                if OpRot(2).torque < OpRot(1).torque
                    OpRot_l = OpRot(2).Op.(strategy);
                else
                    OpRot_u = OpRot(2).Op.(strategy);
                end
            end
            
            if loopCount == Mod.Num.maxIter
            warning('Rotare:Solvers:bemt:NotConverged', ...
                    ['Solution not converged for the coaxial case after reaching the '...
                     'maximum number of iterations (%d iter).\n'], loopCount);
            end
            loopCount = loopCount + 1;
        end
        
    else % If there are more than 2 blades, the function can not support to cancel the total torque
        
         warning('bemt: the total number of rotors in coaxial configuration must be equal to 2.')

         for i = 2:length(OpRot)
                OpRot(i).ElPerf.updateupstreamvel(OpRot(i - 1), Mod); % Update upstream velocity
                bemtsinglerot(OpRot(i), Mod); % Calculate BEMT as usual, with new upstream vel distribution
         end
         
    end
end

function bemtsinglerot(OpRot, Mod)
    % BEMTSINGLEROT  Calculates the BEMT for a single rotor.
    %   This core function solves the BEMT equations using the appropriate solver, then calculates
    %   the forces on each element and the overall performance of the rotor.

    % Determine angles, velocities and forces for each element using the proper solver
    solver = str2func(Mod.solver);
    solver(OpRot, Mod);
            

    % Calculate elemental forces, torque and power
    OpRot.ElPerf.calcforces;
    

    % Calculate overall rotor performance
    OpRot.calcperf;
end
